;; Generated package description from inspector.el  -*- no-byte-compile: t -*-
(define-package "inspector" "0.7" "Tool for inspection of Emacs Lisp objects." '((emacs "27")) :commit "44b3a9aec819b583d76bbbde1421b76c29fe2596" :authors '(("Mariano Montone" . "marianomontone@gmail.com")) :maintainer '("Mariano Montone" . "marianomontone@gmail.com") :keywords '("debugging" "tool" "emacs-lisp" "development") :url "https://github.com/mmontone/emacs-inspector")
